# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import time

import networkx as nx

import minorminer as mm

from .layout import Layout, dnx_layout, p_norm
from .placement import Placement, closest, intersection


def find_embedding(
    S,
    T,
    layout=(p_norm, None),
    placement=closest,
    mm_hint_type="initial_chains",
    return_layouts=False,
    **kwargs
):
    """
    Tries to embed S in T by computing layout-aware chains and passing them to minorminer.find_embedding(). Chains are
    passed as either initial_chains or suspend_chains (see documentation for minorminer.find_embedding to learn more).

    Parameters
    ----------
    S : NetworkX graph or edges data structure (dict, list, ...)
        The graph you are embedding (source) or a NetworkX supported data structure (see to_networkx_graph()).
    T : NetworkX graph or edges data structure (dict, list, ...)
        The graph you are embedding into (target) or a NetworkX supported data structure (see to_networkx_graph()).
    layout : function or [function/dict/Layout, function/dict/Layout] (default None)
        Specifies either a single function to compute the layout for both S and T or a 2-tuple. The 2-tuple either 
        consists of a pair of functions or pre-computed layouts, the first entry in the 2-tuple applies to S while
        the second applies to T. 
        Note: If layout is a single function and T is a dnx_graph, then the function passed in is only applied to S
        and the dnx_layout is applied to T. To run a layout function explicitly on T, pass it in as a 2-tuple; i.e.
        (p_norm, p_norm).
    placement : function or dict (default None)
        If a function, it is the placement algorithm to call; each algorithm uses the layouts of S and T to map the 
        vertices of S to subsets of vertices of T. If it is a dict, it should be a map from the vertices of S to subsets
        of vertices of T.
    mm_hint_type : str (default "initial_chains")
        This is the hint type to tell minorminer.find_embedding(). Supported types are "initial_chains" and 
        "suspend_chains". See minorminer.find_embedding() for more information.
    return_layouts : bool (default False)
        Will return the layout objects of S and T.
    kwargs : dict 
        Keyword arguments are passed to various functions.

    Returns
    -------
    emb : dict
        Output is dependent upon kwargs passed to minorminer, but more or less emb is a mapping from vertices of
        S (keys) to chains in T (values).
    """
    start = time.perf_counter()

    # Parse kwargs
    layout_kwargs, placement_kwargs = _parse_kwargs(kwargs)

    # Compute the layouts
    S_layout, T_layout = _parse_layout_parameter(S, T, layout, layout_kwargs)

    # Compute the placement (i.e. chains)
    S_T_placement = Placement(
        S_layout, T_layout, placement, **placement_kwargs)

    end = time.perf_counter()
    timeout = kwargs.get("timeout")
    if timeout is not None:
        time_remaining = timeout - (end - start)
        if time_remaining <= 0:
            raise TimeoutError(
                "Layout and placement took {}s, you set timeout at {}.".format(end-start, timeout))
        kwargs["timeout"] = time_remaining

    # Run minorminer.find_embedding
    if mm_hint_type == "initial_chains":
        output = mm.find_embedding(
            S, T, initial_chains=S_T_placement, **kwargs)
    elif mm_hint_type == "suspend_chains":
        output = mm.find_embedding(S, T, suspend_chains={
            v: [C] for v, C in S_T_placement.items()}, **kwargs)
    else:
        raise ValueError(
            "Only initial_chains and suspend_chains are supported minorminer hint types.")

    if return_layouts:
        return output, (S_layout, T_layout)
    return output


def _parse_kwargs(kwargs):
    """
    Extract kwargs for layout and placement functions. Leave the remaining ones for minorminer.find_embedding().
    """
    layout_kwargs = {}
    # For the layout object
    if "dim" in kwargs:
        layout_kwargs["dim"] = kwargs.pop("dim")
    if "center" in kwargs:
        layout_kwargs["center"] = kwargs.pop("center")
    if "scale" in kwargs:
        layout_kwargs["scale"] = kwargs.pop("scale")

    placement_kwargs = {}
    # For the placement object
    if "scale_ratio" in kwargs:
        placement_kwargs["scale_ratio"] = kwargs.pop("scale_ratio")
    # For closest strategy
    if "subset_size" in kwargs:
        placement_kwargs["subset_size"] = kwargs.pop("subset_size")
    if "num_neighbors" in kwargs:
        placement_kwargs["num_neighbors"] = kwargs.pop("num_neighbors")

    return layout_kwargs, placement_kwargs


def _parse_layout_parameter(S, T, layout, layout_kwargs):
    """
    Determine what combination of iterable, dict, and function the layout parameter is.
    """
    if nx.utils.iterable(layout):
        try:
            s_layout, t_layout = layout
        except ValueError:
            raise ValueError(
                "layout is expected to be a function or a length-2 iterable")
    else:
        s_layout = t_layout = layout

    # Get a Layout object for S
    if isinstance(s_layout, Layout):
        S_layout = s_layout
    else:
        # Assumes s_layout a callable or implements a mapping interface
        S_layout = Layout(S, layout=s_layout, **layout_kwargs)

    # Get the Layout object for T
    if isinstance(t_layout, Layout):
        T_layout = t_layout
    else:
        # Use the dnx_layout if possible
        if T.graph.get("family") in ("chimera", "pegasus"):
            T_layout = Layout(T, layout=dnx_layout, **layout_kwargs)
        # Assumes t_layout a callable or implements a mapping interface
        else:
            T_layout = Layout(T, layout=t_layout, **layout_kwargs)

    return S_layout, T_layout



# MIT License
# 
# Copyright (c) 2017 Daniel Andersson
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Built-in
import random

# PyPI
import pytest

# Local
import minorminer._rpack as rpack


def test_flat():
    data = [3, 2, 1]
    # Todo: the order should not matter
    assert rpack.group(data, 3) == [[3], [2], [1]]


def test_nr_groups():
    with pytest.raises(ValueError):
        rpack.group([1], -1)
    with pytest.raises(ValueError):
        rpack.group([1], 0)
    data = list(range(1, 100))
    for nr_groups in range(1, 100):
        assert len(rpack.group(data, nr_groups)) == nr_groups


def test_empty_list():
    assert rpack.group([], 3) == [[], [], []]


def test_input_error():
    with pytest.raises(TypeError):
        rpack.group(None, 3)
    with pytest.raises(TypeError):
        rpack.group([1, 2, 3], None)
    with pytest.raises(TypeError):
        rpack.group(None, None)


def test_performance():
    groups = rpack.group(list(range(1_000)), 5)
    assert max(sum(g) for g in groups) == 99_900


def test_swap_complete():
    """Test that no further "swapping" will improve the result"""
    # Get random data
    data = [random.random() for _ in range(500)]
    nr_groups = random.randint(2, 15)
    # Figure out which group is the maximal-sum group
    groups = rpack.group(data, nr_groups)
    group_sums = [sum(g) for g in groups]
    max_duration = max(group_sums)
    group_id_max = group_sums.index(max_duration)
    max_group = groups[group_id_max]

    # For each element in max_group, try to swap element with other
    # elements in the other groups and check if the "max-group" can
    # be improved.
    for i in range(len(max_group)):
        for group_id, group in enumerate(groups):
            if group_id == group_id_max:
                # We don't want to swap within the same group
                continue
            for j in range(len(group)):
                # Swap
                group[j], max_group[i] = max_group[i], group[j]
                # Swap should not make any improvements
                group_sums = [sum(g) for g in groups]
                assert max(group_sums) >= max_duration
                # Undo swap
                group[j], max_group[i] = max_group[i], group[j]



# MIT License
# 
# Copyright (c) 2017 Daniel Andersson
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module docs"""

# Built-in
import collections
from typing import Iterable, Tuple, List

# Local
from . import _rpack


def pack(sizes: Iterable[Tuple[int, int]]) -> List[Tuple[int, int]]:
    """Pack rectangles into an enclosing rectangle with minimal area.

    No rectangles will overlap.

    The GIL is released when C-intensive code is running. Execution
    time increase by number *and* size of rectangles.

    Example::

        # Create a bunch of rectangles (width, height)
        >>> sizes = [(58, 206), (231, 176), (35, 113), (46, 109)]

        # Pack
        >>> positions = rpack.pack(sizes)

        # The result will be a list of (x, y) positions:
        >>> positions
        [(0, 0), (58, 0), (289, 0), (289, 113)]

    :param sizes: Iterable of 2-tuples "(width, height)" of the
        rectangles to pack. **Note: integer values only!**
    :return: List of 2-tuples "(x, y)" of positions of the input
        rectangles. The position refers to the top left corner of the
        rectangle in a coordinate system where the origin is in the top
        left corner.
    """
    return _rpack.pack(sizes)


def group(heights: Iterable[float], nr_groups: int) -> List[List[float]]:
    """Group objects and minimize maximum group sum

    The GIL is released when C-intensive code is running.

    :param heights: Iterable of objects which implements float(object).
    :param nr_groups: Nr of groups.
    :type nr_groups: int
    :return: List of lists. Each list represents a group and contains
        the groups objects.
    """
    return _rpack.group(heights, nr_groups)


_R = collections.namedtuple('R', 'width height x y')


def enclosing_size(sizes, positions):
    """Return enclosing size of rectangles having sizes and positions.

    Useful to compute the enclosing size of the output of
    :py:func:`rpack.pack`.

    Example::

        >>> sizes = [(58, 206), (231, 176), (35, 113), (46, 109)]
        >>> positions = rpack.pack(sizes)

        >>> enclosing_size(sizes, positions)
        (335, 222)


    """
    rectangles = [_R(*size, *pos) for size, pos in zip(sizes, positions)]
    width = max(r.width + r.x for r in rectangles)
    height = max(r.height + r.y for r in rectangles)
    return width, height



# Copyright 2020 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import random
from collections import Counter, abc, defaultdict

import networkx as nx
import numpy as np
from scipy import spatial

from . import layout


def intersection(S_layout, T_layout, **kwargs):
    """
    Map each vertex of S to its nearest row/column intersection qubit in T (T must be a D-Wave hardware graph).
    Note: This will modifiy S_layout. 

    Parameters
    ----------
    S_layout : layout.Layout
        A layout for S; i.e. a map from S to R^d.
    T_layout : layout.Layout
        A layout for T; i.e. a map from T to R^d.
    scale_ratio : float (default None)
        If None, S_layout is not scaled. Otherwise, S_layout is scaled to scale_ratio*T_layout.scale.

    Returns
    -------
    placement : dict
        A mapping from vertices of S (keys) to vertices of T (values).
    """
    # Extract the target graph
    T = T_layout.G

    # Currently only implemented for 2d chimera
    if T.graph.get("family") not in ("chimera", "pegasus"):
        raise NotImplementedError(
            "This strategy is currently only implemented for Chimera.")

    # Bin vertices of S and T into a grid graph G
    G = _intersection_binning(S_layout, T)

    placement = {}
    for _, data in G.nodes(data=True):
        for v in data["variables"]:
            placement[v] = data["qubits"]

    return placement


def _intersection_binning(S_layout, T):
    """
    Map the vertices of S to the "intersection graph" of T. This modifies the grid graph G by assigning vertices 
    from S and T to vertices of G.

    Parameters
    ----------
    S_layout : layout.Layout
        A layout for S; i.e. a map from S to R^d.
    T : networkx.Graph
        The target graph to embed S in.
    scale_ratio : float (default None)
        If None, S_layout is not scaled. Otherwise, S_layout is scaled to scale_ratio*T_layout.scale.

    Returns
    -------
    G : networkx.Graph
        A grid graph. Each vertex of G contains data attributes "variables" and "qubits", that is, respectively 
        vertices of S and T assigned to that vertex.  
    """
    # Scale the layout so that for each unit-cell edge, we have an integer point.
    m, n, t = T.graph["rows"], T.graph["columns"], T.graph["tile"]

    # --- Make the "intersection graph" of the dnx_graph
    # Grid points correspond to intersection rows and columns of the dnx_graph
    G = nx.grid_2d_graph(t*n, t*m)

    # Determine the scale for putting things in the positive quadrant
    scale = (t*min(n, m) - 1)/2

    # Get the row, column mappings for the dnx graph
    lattice_mapping = _lookup_intersection_coordinates(T)

    # Less efficient, but more readable to initialize all at once
    for v in G:
        G.nodes[v]["qubits"] = set()
        G.nodes[v]["variables"] = set()

    # Add qubits (vertices of T) to grid points
    for int_point, Q in lattice_mapping.items():
        G.nodes[int_point]["qubits"] |= Q

    # --- Map the S_layout to the grid
    # "Zoom in" on layout_S so that the integer points are better represented
    zoom_scale = S_layout.scale*t
    if zoom_scale < scale:
        S_layout.scale = zoom_scale
    else:
        S_layout.scale = scale

    # Center to the positive orthant
    S_layout.center = 2*(scale, )

    # Add "variables" (vertices from S) to grid points too
    for v, pos in S_layout.items():
        grid_point = tuple(int(x) for x in np.round(pos))
        G.nodes[grid_point]["variables"].add(v)

    return G


def _lookup_intersection_coordinates(G):
    """
    For a dwave_networkx graph G, this returns a dictionary mapping the lattice points to sets of vertices of G. 
        - Chimera: Each lattice point corresponds to the 2 qubits intersecting at that point.
        - Pegasus: Not Implemented
    """
    graph_data = G.graph
    family = graph_data.get("family")

    if family == "chimera":
        t = graph_data.get("tile")
        intersection_points = defaultdict(set)
        if graph_data["labels"] == "coordinate":
            for v in G:
                _chimera_all_intersection_points(intersection_points, v, t, *v)

        elif graph_data["data"]:
            for v, d in G.nodes(data=True):
                _chimera_all_intersection_points(
                    intersection_points, v, t, *d["chimera_index"])

        else:
            raise NotImplementedError("Please pass in a Chimera graph created"
                " with an optional parameter 'data=True' or 'coordinates=True'")

        return intersection_points

    elif family == "pegasus":
        offsets = [graph_data['vertical_offsets'],
                   graph_data['horizontal_offsets']]

        intersection_points = defaultdict(set)
        if graph_data["labels"] == "coordinate":
            for v in G:
                _pegasus_all_intersection_points(intersection_points, offsets,
                                                 v, *v)
        elif graph_data["data"]:
            for v, d in G.nodes(data=True):
                _pegasus_all_intersection_points(intersection_points, offsets,
                                                 v, *d["pegasus_index"])
        else:
            raise NotImplementedError("Please pass in a Pegasus graph created"
                " with an optional parameter 'data=True' or 'coordinates=True'")

        return intersection_points



def _chimera_all_intersection_points(intersection_points, v, t, i, j, u, k):
    """
    Given a coordinate vertex, v = (i, j, u, k), of a Chimera with tile, t, get all intersection points it is in.
    """
    # If you're a row vertex, you go in all grid points of your row intersecting columns in your unit tile
    if u == 1:
        row = i*t + k
        for kk in range(t):
            col = j*t + kk
            intersection_points[(col, row)].add(v)

    # Sameish for a column vertex.
    elif u == 0:
        col = j*t + k
        for kk in range(t):
            row = i*t + kk
            intersection_points[(col, row)].add(v)

def _pegasus_all_intersection_points(intersection_points, offsets, v, u, w, k, z):
    """
    Given a coordinate vertex, v = (u, w, k, z), of a Pegasus graph with offsets
    `offsets`, get all intersection points it is in.
    """
    # Each horizontal qubit spans twelve grid-points in the row 12w+k
    if u == 1:
        row = 12*w + k
        col_0 = 12*z + offsets[u][k]
        for kk in range(12):
            intersection_points[(col_0 + kk, row)].add(v)

    # Sameish for a column vertex.
    elif u == 0:
        col = 12*w + k
        row_0 = 12*z + offsets[u][k]
        for kk in range(12):
            intersection_points[(col, row_0 + kk)].add(v)


def closest(S_layout, T_layout, subset_size=(1, 1), num_neighbors=1, **kwargs):
    """
    Maps vertices of S to the closest vertices of T as given by S_layout and T_layout. i.e. For each vertex u in
    S_layout and each vertex v in T_layout, map u to the v with minimum Euclidean distance (||u - v||_2).

    Parameters
    ----------
    S_layout : layout.Layout
        A layout for S; i.e. a map from S to R^d.
    T_layout : layout.Layout
        A layout for T; i.e. a map from T to R^d.
    subset_size : tuple (default (1, 1))
        A lower (subset_size[0]) and upper (subset_size[1]) bound on the size of subets of T that will be considered
        when mapping vertices of S.
    num_neighbors : int (default 1)
        The number of closest neighbors to query from the KDTree--the neighbor with minimium overlap is chosen.
        Increasing this reduces overlap, but increases runtime.

    Returns
    -------
    placement : dict
        A mapping from vertices of S (keys) to subsets of vertices of T (values).
    """
    # Extract the target graph
    T = T_layout.G

    # A new layout for subsets of T.
    T_subgraph_layout = {}

    # Get connected subgraphs to consider mapping to
    T_subgraphs = _get_connected_subgraphs(T, subset_size[1])

    # Calculate the barycenter (centroid) of each subset
    for k in range(subset_size[0], subset_size[1]+1):
        if k == 1:
            for subgraph in T_subgraphs[k]:
                v, = subgraph  # Unpack the subgraph of size 1
                T_subgraph_layout[subgraph] = T_layout[v]
        else:
            for subgraph in T_subgraphs[k]:
                T_subgraph_layout[subgraph] = np.mean(
                    np.array([T_layout[v] for v in subgraph]), axis=0)

    # Use scipy's KDTree to solve the nearest neighbor problem.
    # This requires a few lookup tables
    T_subset_lookup = {tuple(p): V for V, p in T_subgraph_layout.items()}
    layout_points = [tuple(p) for p in T_subgraph_layout.values()]
    overlap_counter = Counter()

    try:
        tree = spatial.KDTree(layout_points)  # This fails for the empty graph
    except ValueError:
        pass

    placement = {}
    for u, u_pos in S_layout.items():
        distances, v_indices = tree.query([u_pos], num_neighbors)

        # KDTree.query either returns a (num_neighbors, ) shaped arrays if num_neighbors == 1
        # or (1, num_neighbors) shaped arrays if num_neighbors != 1
        if num_neighbors != 1:
            v_indices = v_indices[0]
            distances = distances[0]

        placement[u] = _minimize_overlap(
            distances, v_indices, T_subset_lookup, layout_points, overlap_counter)

    return placement


def _get_connected_subgraphs(G, k, single_set=False):
    """
    Finds all connectected subgraphs S of G within a given subset_size.

    Parameters
    ----------
    G : networkx graph
        The graph you want to find all connected subgraphs of.
    k : int
        An upper bound of the size of connected subgraphs to find.

    Returns
    -------
    connected_subgraphs : dict
        The dictionary is keyed by size of subgraph and each value is a set containing
        frozensets of vertices that comprise the connected subgraphs.
        {
            1: { {v_1}, {v_2}, ... },
            2: { {v_1, v_2}, {v_1, v_3}, ... },
            ...,
            k: { {v_1, v_2, ..., v_m}, ... }
        }
    """
    connected_subgraphs = defaultdict(set)
    connected_subgraphs[1] = {frozenset((v,)) for v in G}

    for i in range(1, min(k, len(G))):
        # Iterate over the previous set of connected subgraphs.
        for X in connected_subgraphs[i]:
            # For each vertex in the set, iterate over its neighbors.
            for v in X:
                for u in G.neighbors(v):
                    connected_subgraphs[i + 1].add(X.union({u}))

    return connected_subgraphs


def _minimize_overlap(distances, v_indices, T_subset_lookup, layout_points, overlap_counter):
    """
    A greedy penalty-type model for choosing nonoverlapping chains.
    """
    subsets = {}
    for i, d in zip(v_indices, distances):
        subset = T_subset_lookup[layout_points[i]]
        subsets[subset] = d + sum(10**overlap_counter[v] for v in subset)

    cheapest_subset = min(subsets, key=subsets.get)
    overlap_counter.update(cheapest_subset)
    return cheapest_subset


class Placement(abc.MutableMapping):
    def __init__(
        self,
        S_layout,
        T_layout,
        placement=None,
        scale_ratio=None,
        **kwargs
    ):
        """
        Compute a placement of S in T, i.e., map V(S) to 2^{V(T)}.

        Parameters
        ----------
        S_layout : layout.Layout
            A layout for S; i.e. a map from S to R^d.
        T_layout : layout.Layout
            A layout for T; i.e. a map from T to R^d.
        placement : dict or function (default None)
            If a dict, this specifies a pre-computed placement for S in T. If a function, the function is called on
            S_layout and T_layout `placement(S_layout, T_layout)` and should return a placement of S in T. If None,
            a random placement of S in T is selected.
        scale_ratio : float (default None)
            If None, S_layout is not scaled. Otherwise, S_layout is scaled to scale_ratio*T_layout.scale.
        kwargs : dict
            Keyword arguments are given to placement if it is a function.
        """
        self.S_layout = _parse_layout(S_layout)
        self.T_layout = _parse_layout(T_layout)

        # Layout dimensions should match
        if self.S_layout.dim != self.T_layout.dim:
            raise ValueError(
                "S_layout has dimension {} but T_layout has dimension {}. These must match.".format(
                    self.S_layout.dim, self.T_layout.dim)
            )

        # Scale S if S_layout is bigger than T_layout
        if self.S_layout.scale > self.T_layout.scale:
            self.S_layout.scale = self.T_layout.scale
        # Or scale S to the user specified scale
        elif scale_ratio:
            self.S_layout.scale = scale_ratio*self.T_layout.scale

        if placement is None:
            self.placement = closest(self.S_layout, self.T_layout)
        elif callable(placement):
            self.placement = placement(self.S_layout, self.T_layout, **kwargs)
        else:
            self.placement = placement

    # The class should behave like a dictionary
    def __iter__(self):
        """
        Iterate through the keys of the dictionary placement.
        """
        yield from self.placement

    def __getitem__(self, key):
        """
        Get the placement value at the key vertex.
        """
        return self.placement[key]

    def __setitem__(self, key, value):
        """
        Set the placement value at the key vertex.
        """
        self.placement[key] = value

    def __delitem__(self, key):
        """
        Delete the placement value at the key vertex.
        """
        del self.placement[key]

    def __repr__(self):
        """
        Use the placement's dictionary representation.
        """
        return repr(self.placement)

    def __len__(self):
        """
        The length of a placement is the length of the placement dictionary.
        """
        return len(self.placement)


def _parse_layout(G_layout):
    """
    Ensures a Layout object was passed in and makes a copy to save in the Placement object.
    """
    if isinstance(G_layout, layout.Layout):
        return layout.Layout(G_layout.G, G_layout.layout)

    else:
        raise TypeError(
            "If you want to pass in a precomputed layout mapping, please create a Layout object; Layout(G, layout).")



# MIT License
# 
# Copyright (c) 2017 Daniel Andersson
# 
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Built-in
from collections import namedtuple
import random

# PyPI
import pytest

# Local
import minorminer._rpack as rpack

R = namedtuple('R', 'width height x y')


def enclosing_size(sizes, positions):
    """Return enclosing size of rectangles having sizes and positions"""
    rectangles = [R(*size, *pos) for size, pos in zip(sizes, positions)]
    width = max(r.width + r.x for r in rectangles)
    height = max(r.height + r.y for r in rectangles)
    return width, height


def test_enclosing_size():
    """Test enclosing size helper function"""
    sizes = [(3, 5), (1, 1), (1, 1)]
    positions = [(0, 0), (3, 0), (0, 5)]
    width, height = enclosing_size(sizes, positions)
    assert width == 4
    assert height == 6

def test_origin():
    """Single rectangle should be positioned in origin"""
    rectangles = [(3, 5)]
    positions = [(0, 0)]
    assert rpack.pack(rectangles) == positions

def test_perfect_pack():
    """Pack rectangles to perfect rectangle

    Like this::

        aaa  bb  cc  -->  aaabb
        aaa  bb           aaabb
        aaa               aaacc
    """
    rectangles = [(3, 3), (2, 2), (2, 1)]
    positions = [(0, 0), (3, 0), (3, 2)]
    assert rpack.pack(rectangles) == positions

def test_basic_pack():
    """Single rectangle should be positioned in origin"""
    sizes = [(2, 2), (2, 2), (2, 2), (3, 3)]
    positions = rpack.pack(sizes)
    width, height = enclosing_size(sizes, positions)
    assert width*height == 25

def test_medium_pack():
    sizes = [(i, i) for i in range(20, 1, -1)]
    positions = rpack.pack(sizes)
    width, height = enclosing_size(sizes, positions)
    assert width*height <= 3045

def test_no_overlap():
    """Make sure no rectangles overlap"""
    random.seed(123)
    rectangles = [(random.randint(50, 100), random.randint(50, 100))
                  for _ in range(40)]
    positions = rpack.pack(rectangles)
    for i, ((x1, y1), (w1, h1)) in enumerate(zip(positions, rectangles)):
        for j, ((x2, y2), (w2, h2)) in enumerate(zip(positions, rectangles)):
            if i != j:
                disjoint_in_x = (x1 + w1 <= x2 or x2 + w2 <= x1)
                disjoint_in_y = (y1 + h1 <= y2 or y2 + h2 <= y1)
                assert disjoint_in_x or disjoint_in_y

# TEST INPUT
# ==========

def test_empty():
    """Empty input should give empty output"""
    rectangles = []
    positions = []
    assert rpack.pack(rectangles) == positions

def test_zero():
    """Zero number should raise ValueError"""
    with pytest.raises(ValueError):
        rpack.pack([(0, 0)])

def test_negative():
    """Negative number should raise ValueError"""
    with pytest.raises(ValueError):
        rpack.pack([(3, -5)])
    with pytest.raises(ValueError):
        rpack.pack([(-3, 5)])
    with pytest.raises(ValueError):
        rpack.pack([(-3, -5)])

def test_not_iterable():
    """Non-number should raise ValueError"""
    with pytest.raises(TypeError):
        rpack.pack(None)

def test_not_height_width():
    """Non-number should raise ValueError"""
    with pytest.raises(TypeError):
        rpack.pack([None])

def test_not_integers():
    """Non-number should raise ValueError"""
    with pytest.raises(TypeError):
        rpack.pack([('garnet', 9)])
    with pytest.raises(TypeError):
        rpack.pack([(9, 'alexandros')])

def test_floats():
    with pytest.raises(TypeError):
        rpack.pack([[1.99, 1.99]])
